// see CSvgGeometryBuilder.h
#define _INOVAE_CURVETYPE_LINE      2
#define _INOVAE_CURVETYPE_QUADRATIC 3
#define _INOVAE_CURVETYPE_CUBIC     4
#define _INOVAE_CURVETYPE_RATIONAL	19

#define _INOVAE_UI_FRAG_SIZE        2

#define _INOVAE_UI_ARCW1_DEFAULT	-1

#define _INOVAE_UI_RASTERIZER_WORKGROUP_SIZE 256
#define _INOVAE_UI_GEN_STENCIL_WORKGROUP_SIZE 1024
#define _INOVAE_UI_GEN_STENCIL_BLOCK_BOUNDARY_WORKGROUP_SIZE 256

#define _INOVAE_UI_SAMPLE_PACKING_SCALE 0.5f

#define _INOVAE_FILL_MODE_NON_ZERO 0
#define _INOVAE_FILL_MODE_EVEN_ODD 1

struct BlockSortStencilDispatchInfo
{
	// xyz = dispatch info, w = frag count
	int4 m_genStencilFragDispatchInfo;

	// xyz = dispatch info, w = numBlocks
	int4 m_binarySearchFragDispatchInfo;

	// xyz = dispatch info
	int4 m_blockSortFragDispatchInfo;
};

bool PathVertexIsInvisible(uint4 flags0, uint4 flags1)
{
	uint lower = (flags0.x & 0xff) | ((flags0.y & 0xff) << 8) | ((flags0.z & 0xff) << 16) | ((flags0.w & 0xff) << 24);
	uint upper = (flags1.x & 0xff) | ((flags1.y & 0xff) << 8) | ((flags1.z & 0xff) << 16) | ((flags1.w & 0xff) << 24);

	return ((lower & 0x01000000) == 0 && (upper & 0x01010101) == 0)
		|| ((lower & 0x01000101) == 0 && (upper & 0x00010100) == 0)
		|| ((lower & 0x01010101) == 0 && (upper & 0x00000001) == 0)
		|| ((lower & 0x00010100) == 0 && (upper & 0x01010001) == 0);

	//bvec4 result = !(flags0 == uint4(0, 0, 0, 0) && flags1 == uint4(1, 1, 1, 1))
	//    || !(flags0 == uint4(1, 1, 0, 1) && flags1 == uint4(0, 1, 1, 0))
	//    || !(flags0 == uint4(1, 1, 1, 1) && flags1 == uint4(0, 0, 0, 0))
	//    || !(flags0 == uint4(0, 1, 1, 0) && flags1 == uint4(1, 0, 1, 1));

	//return result.x && result.y && result.z && result.w;
}

bool PathVertexIsVisible(uint4 flags0, uint4 flags1)
{
	return !PathVertexIsInvisible(flags0, flags1);
}

uint PointLocation(uint i, uint threadIdx)
{
	return i * _INOVAE_UI_RASTERIZER_WORKGROUP_SIZE + threadIdx;
}

uint T1QueueLocation(uint i, uint threadIdx)
{
	return i * _INOVAE_UI_RASTERIZER_WORKGROUP_SIZE + threadIdx;
}

float interpolateGeneralCurve(
	uint mode, float arcw1, float v0, float v1, float v2, float v3, float t)
{
	switch(mode)
	{
	case _INOVAE_CURVETYPE_LINE:
	{
		return lerp(v0, v1, t);
	}
	case _INOVAE_CURVETYPE_QUADRATIC:
	{
		float lv0 = lerp(v0, v1, t);
		float lv1 = lerp(v1, v2, t);

		return lerp(lv0, lv1, t);
	}
	case _INOVAE_CURVETYPE_CUBIC:
	{
		float iv0 = v0;
		float iv1 = v1;
		float iv2 = v2;
		float iv3 = v3;

		float qv0 = lerp(iv0, iv1, t);
		float qv1 = lerp(iv1, iv2, t);
		float qv2 = lerp(iv2, iv3, t);

		float lv0 = lerp(qv0, qv1, t);
		float lv1 = lerp(qv1, qv2, t);

		return lerp(lv0, lv1, t);
	}
	/*else if(mode == CT_Rational)
	{
	float iv0 = v0;
	float iv1 = v1 *arcw1;
	float iv2 = v2;

	auto iw = arcw1;

	float lv0 = ::lerp(iv0, iv1, t);
	float lv1 = ::lerp(iv1, iv2, t);
	float v = ::lerp(lv0, lv1, t);
	float w = (1 - t)*(1 - t) + 2.f*(1 - t)*t*iw + t*t;
	v *= safeRcp(w);
	return v;
	}*/
	}

	return 0.0f;
}

int d_float2int_rd(float x, int D)
{
	if(D == 2)
	{
		return int(floor(x)) & 0xFFFFFFFE;
	}

	if(D == 4)
	{
		return int(floor(x)) & 0xFFFFFFFC;
	}

	return int(floor(x / D)) * D;
}

int d_round_down(int x, int D)
{
	if(D == 2)
	{
		return x & 0xFFFFFFFE;
	}

	if(D == 4)
	{
		return x & 0xFFFFFFFC;
	}

	return x;
}

float safeRcp(float a)
{
	return a != 0.0f ? (1.0f / a) : 0.0f;
}

void solveQuadEquation(float a, float b, float c, out float r0, out float r1)
{
	if(a == 0)
	{
		float x = -c / b;

		r0 = x;
		r1 = x;

		return;
	}

	float A = a;
	float B = b * 0.5;
	float C = c;
	float2 tc;

	float R = B * B - A * C;

	if(R > 0.0f)
	{
		float SR = sqrt(R);

		if(B > 0.0f)
		{
			float TB = B + SR;
			tc = float2(-C / TB, -TB / A);
		}
		else // B<0
		{
			float TB = -B + SR;
			tc = float2(TB / A, C / TB);
		}
	}
	else
	{
		tc = float2(0.0f, 0.0f);
	}

	r0 = tc.x;
	r1 = tc.y;
}

float solveQuadBezierQM(float iA, float iB, float iC, float iD, float t0, float t1)
{
	// A*(1-t)^2 + 2*B*t*(1-t) + C*t^2 = D

	float a = iA + (-2.0f) * iB + iC;
	float b = -2.0f * iA + 2.0f * iB;
	float c = iA - iD;

	float r0 = 0.0f;
	float r1 = 0.0f;

	solveQuadEquation(a, b, c, r0, r1);

	if(t0 <= r0 && r0 <= t1)
	{
		return r0;
	}

	if(t0 <= r1 && r1 <= t1)
	{
		return r1;
	}

	return t0;
}

float2 bezier_curve_point(const float2 v0, const float2 v1, float t)
{
	return lerp(v0, v1, float2(t, t));
}

float2 bezier_curve_point(const float2 v0, const float2 v1, const float2 v2, float t)
{
	float2 t2 = float2(t, t);

	float2 lv0 = lerp(v0, v1, t2);
	float2 lv1 = lerp(v1, v2, t2);

	return lerp(lv0, lv1, t2);
}

float2 bezier_curve_point(const float2 iv0, const float2 iv1, const float2 iv2, const float2 iv3, float t)
{
	float2 t2 = float2(t, t);

	float2 qv0 = lerp(iv0, iv1, t2);
	float2 qv1 = lerp(iv1, iv2, t2);
	float2 qv2 = lerp(iv2, iv3, t2);

	float2 lv0 = lerp(qv0, qv1, t2);
	float2 lv1 = lerp(qv1, qv2, t2);

	return lerp(lv0, lv1, t2);
}

float2 bezier_curve_point(const float2 iv0, const float2 iv1, const float2 iv2, const float iw, float t)
{
	float2 t2 = float2(t, t);

	float2 lv0 = lerp(iv0, iv1, t2);
	float2 lv1 = lerp(iv1, iv2, t2);
	float2 v = lerp(lv0, lv1, t2);

	float w = (1 - t) * (1 - t) + 2.f * (1 - t) * t * iw + t * t;
	v *= safeRcp(w);

	return v;
}

float2 d_curve_point_Linear(const float2 iv0, const float2 iv1, const float2 iv2, const float2 iv3, const float iw, float t)
{
	return lerp(iv0, iv1, float2(t, t));
}

float2 d_curve_point_Quadratic(const float2 iv0, const float2 iv1, const float2 iv2, const float2 iv3, const float iw, float t)
{
	return bezier_curve_point(iv0, iv1, iv2, t);
}

float2 d_curve_point_Cubic(const float2 iv0, const float2 iv1, const float2 iv2, const float2 iv3, const float iw, float t)
{
	return bezier_curve_point(iv0, iv1, iv2, iv3, t);
}

float2 d_curve_point_Rational(const float2 iv0, const float2 iv1, const float2 iv2, const float2 iv3, const float iw, float t)
{
	return bezier_curve_point(iv0, iv1, iv2, iw, t);
}

void shiftRight64(inout uint low, inout uint high, int shiftAmount)
{
	if(shiftAmount > 32)
	{
		high = 0;
		low >>= (shiftAmount - 32);
	}
	else
	{
		high >>= shiftAmount;
	}
}

int4 fetchVerticalTable(float y, int numSamples)
{
	if(numSamples == 8)
	{
		int mask = 0xffff;
		mask >>= (16 - int(round(y * 16)));
		mask |= mask << 16;

		return int4(mask, 0, 0, 0);
	}

	// uint64_t mask = 0xFFFFFFFFFFFFFFFFLL;
	// mask >>= (64 - (int)round(y * 64));
	// int m0 = mask;
	// int m1 = mask >> 32;
	// return make_int4(m0, m1, m0, m1);

	int shiftAmt = 64 - int(round(y * 64));

	uint m0 = 0xffffffff;
	uint m1 = 0xffffffff;

	shiftRight64(m0, m1, shiftAmt);

	return int4(int(m0), int(m1), int(m0), int(m1));
}

int4 fetchVerticalTable2(float y0, float y1, int numSamples)
{
	//int4 m0, m1;
	//m0 = fetchVerticalTable(y0);
	//m1 = fetchVerticalTable(y1);
	//return make_int4(m0.x ^ m1.x, m0.y ^ m1.y, m0.z ^ m1.z, m0.w ^ m1.w);

	if(numSamples == 8)
	{
		//int shift_0 = (16 - (int)round(y0 * 16));
		//int shift_1 = (16 - (int)round(y1 * 16));
		int shift_0 = (16 - int(y0 * 16 + 0.5));
		int shift_1 = (16 - int(y1 * 16 + 0.5));
		shift_0 = max(shift_0, 0);
		shift_1 = max(shift_1, 0);

		int mask = 0xffff;
		mask = (mask >> shift_0) ^ (mask >> shift_1);
		mask |= mask << 16;

		return int4(mask, 0, 0, 0);
	}

	//QM: to test - shift by 64 may not work
	//int shift_0 = (64 - (int)round(y0 * 64));
	//int shift_1 = (64 - (int)round(y1 * 64));
	int shift_0 = (64 - int(y0 * 64 + 0.5));
	int shift_1 = (64 - int(y1 * 64 + 0.5));

	shift_0 = max(shift_0, 0);
	shift_1 = max(shift_1, 0);

	//uint64_t mask = 0xFFFFFFFFFFFFFFFFLL;
	//mask = (mask >> shift_0) ^ (mask >> shift_1);

	//int m0 = mask;
	//int m1 = mask >> 32;

	uint low0 = 0xffffffff;
	uint high0 = 0xffffffff;
	uint low1 = 0xffffffff;
	uint high1 = 0xffffffff;

	shiftRight64(low0, high0, shift_0);
	shiftRight64(low1, high1, shift_1);

	int m0 = int(low0 ^ low1);
	int m1 = int(high0 ^ high1);

	return int4(m0, m1, m0, m1);
}

float2 fetchPixelTable(float2 P0, float2 P1, int numSamples)
{
	//if (P1.x < P0.x) { swap(P0, P1); }
	if(P1.y < P0.y)
	{
		// swap
		float2 tmp = P0;

		P0 = P1;
		P1 = tmp;
	}

	float2 D = P1 - P0;
	float2 N = normalize(float2(D.y, -D.x));
	float C = dot(N, P0);

	//translate origin to (0.5,0.5)
	C -= 0.5f * (N.x + N.y);

	float C_sign = asfloat((asint(C) & 0x80000000) + 0x3f800000);

	/*if (C<0.f) {
	N = -N;
	C = -C;
	}*/

	//reverse the distance
	float2 N_rev = max(1.f - C * C_sign * _INOVAE_UI_SAMPLE_PACKING_SCALE, 0.f) * C_sign * N;
	float2 uv = N_rev * 0.5f + float2(0.5f, 0.5f);

	return uv;

	//if (N_SAMPLES == 8)
	//{
	//    return int4(tex2D(tex_pixel_table8, uv.x, uv.y), 0, 0, 0);
	//}
	//else
	//{
	//    return tex2D(tex_pixel_table32, uv.x, uv.y);
	//}
}

//layout(std140, binding = 0) uniform cbVertices
//{
//	float2 g_vertices[];
//};

/************************* Set 0 *************************/

#if _INOVAE_D3D12

// x = x, y = y, z = primitive ID encoded as float
Buffer<float3> g_vertices : register(t0, space1);

RWBuffer<float2> g_transformedVertices : register(u0, space1);

RWBuffer<uint> g_pathVisibility : register(u1, space1);

#else

// x = x, y = y, z = primitive ID encoded as float
Buffer<float3> g_vertices : register(t0);

RWBuffer<float2> g_transformedVertices : register(u0);

RWBuffer<uint> g_pathVisibility : register(u1);

#endif

/************************* Push Constants *************************/

cbuffer cbPushConstants : register(b0)
{
	float4x4 g_matWorldViewProj;

	//x = vertex count
	uint4 g_counts;

	// x = viewport width, y = viewport height, z = viewport depth, w = cutoff
	float4 g_dimensions;
};

/************************* Code *************************/

// Dispatch with ((numVerts + (local_size_x - 1)) / local_size_x) work groups
[numthreads(_INOVAE_UI_RASTERIZER_WORKGROUP_SIZE, 1, 1)]
void main(uint3 globalInvocationID : SV_DispatchThreadID)
{
	if(globalInvocationID.x >= g_counts.x)
	{
		return;
	}

	float3 vertex = g_vertices[globalInvocationID.x];
	uint pathID = asuint(vertex.z);
	float4 transformedVertex = mul(float4(vertex.xy, 0.0, 1.0), g_matWorldViewProj);

	transformedVertex.xyz /= transformedVertex.w;

	if(transformedVertex.z < g_dimensions.w)
	{
		transformedVertex.z = g_dimensions.w;
	}

	float2 halfDim = g_dimensions.xy * 0.5;

	transformedVertex.xy = (transformedVertex.xy - halfDim) * g_dimensions.z / (transformedVertex.z + g_dimensions.z) + halfDim;

	// path visible flag:
	//
	//         y > height
	//
	//         5 | 6 | 7
	// x < 0   3 | x | 4  x > width
	//         0 | 1 | 2 
	//
	//           y < 0

	uint x_flag = transformedVertex.x < 0 ? 0 : (transformedVertex.x <= g_dimensions.x ? 1 : 2);
	uint y_flag = transformedVertex.y < 0 ? 0 : (transformedVertex.y <= g_dimensions.y ? 1 : 2);

	// I really wish GLSL supported bytes, so much wasted space
	switch((y_flag << 4) | x_flag)
	{
	case 0x00: g_pathVisibility[pathID * 8] = 1; break;
	case 0x01: g_pathVisibility[pathID * 8 + 1] = 1; break;
	case 0x02: g_pathVisibility[pathID * 8 + 2] = 1; break;

	case 0x10: g_pathVisibility[pathID * 8 + 3] = 1; break;
	case 0x11: g_pathVisibility[pathID * 8] = 1; g_pathVisibility[pathID * 8 + 7] = 1; break;
	case 0x12: g_pathVisibility[pathID * 8 + 4] = 1; break;

	case 0x20: g_pathVisibility[pathID * 8 + 5] = 1; break;
	case 0x21: g_pathVisibility[pathID * 8 + 6] = 1; break;
	case 0x22: g_pathVisibility[pathID * 8 + 7] = 1; break;

	default: break;
	}

	if(floor(transformedVertex.x) == transformedVertex.x)
	{
		transformedVertex.x = transformedVertex.x - 1.0 / 256.0;
	}

	g_transformedVertices[globalInvocationID.x] = transformedVertex.xy;
}